;; Generated package description from cpio-mode.el  -*- no-byte-compile: t -*-
(define-package "cpio-mode" "0.17.0.20201115.95820" "Handle cpio archives in the style of dired." '((emacs "24.5")) :url "https://elpa.gnu.org/packages/cpio-mode.html" :authors '(("Douglas Lewan" . "d.lewan2000@gmail.com")) :maintainer '("Douglas Lewan" . "d.lewan2000@gmail.com") :keywords '("files"))
